import json
import copy
import six
from montague.interfaces import IConfigLoader, IConfigLoaderFactory
from montague.structs import LoadableConfig, DEFAULT
from montague.vendor import reify
from zope.interface import directlyProvides, implementer


@implementer(IConfigLoader)
class JSONConfigLoader(object):
    """This is a sample config loader. It uses a structural
       convention that makes more sense for JSON; the root object
       contains 'application' and 'server' keys, which each contain keys
       for the respective items. It has basically no error handling
       and is kind of inefficient."""
    directlyProvides(IConfigLoaderFactory)

    def __init__(self, path):
        self.path = path

    @reify
    def _config(self):
        return json.load(open(self.path))

    def config(self):
        config = {}
        for section, vals in six.iteritems(self._config):
            if 'main' in vals:
                vals = copy.deepcopy(vals)
                vals[DEFAULT] = vals.pop('main')
            config[section] = vals
        return config

    def app_config(self, name):
        if name is DEFAULT:
            name = 'main'
        # Obviously this will throw a KeyError if the config isn't there.
        # A real implementation would have error handling here.
        config = self._config['application'][name]
        return LoadableConfig.app(name=name, config=config, global_config={})

    def server_config(self, name):
        if name is DEFAULT:
            name = 'main'
        config = self._config['server'][name]
        return LoadableConfig.server(name=name, config=config, global_config={})

    def filter_config(self, name):
        # Not covered in tests yet.
        raise NotImplementedError



import os
import sys
import shutil

from pkg_resources import working_set, require
import pytest


@pytest.fixture(scope='session')
def fakeapp():
    test_dir = os.path.dirname(__file__)
    egg_info_dir = os.path.join(
        test_dir, 'fake_packages', 'FakeApp.egg', 'EGG-INFO')
    info_dir = os.path.join(
        test_dir, 'fake_packages', 'FakeApp.egg', 'FakeApp.egg-info-dir')
    if not os.path.exists(egg_info_dir):
        try:
            os.symlink(info_dir, egg_info_dir)
        except:
            shutil.copytree(info_dir, egg_info_dir)

    sys.path.append(os.path.dirname(egg_info_dir))

    working_set.add_entry(os.path.dirname(egg_info_dir))
    require('FakeApp')
    import fakeapp.apps
    return fakeapp



from setuptools import setup, find_packages

setup(
    name="FakeApp",
    version="1.0",
    packages=find_packages(),
    entry_points={
      'paste.app_factory': """
      basic_app=fakeapp.apps:make_basic_app
      other=fakeapp.apps:make_basic_app2
      configed=fakeapp.configapps:SimpleApp.make_app
      """,
      'paste.composit_factory': """
      remote_addr=fakeapp.apps:make_remote_addr
      """,
      'paste.filter_factory': """
      caps=fakeapp.apps:make_cap_filter
      """,
      'paste.filter_app_factory': """
      caps2=fakeapp.apps:CapFilter
      """,
      'paste.server_factory': """
      server_factory=fakeapp.servers.make_server_factory
      """,
      'paste.server_runner': """
      server_runner=fakeapp.servers.make_server_runner
      """,
      'montague.config_loader': """
      json=fakeapp.config:JSONConfigLoader
      """,
      },
    )



from pprint import pprint
import sys

import pkg_resources

from paste.deploy.compat import print_


def test_load_package():
    print_('Path:')
    pprint(sys.path)
    print_(pkg_resources.require('FakeApp'))



from __future__ import absolute_import

from montague.interfaces import IConfigLoader, IConfigLoaderFactory
from zope.interface import directlyProvides, implementer
from characteristic import attributes
from montague.structs import DEFAULT, LoadableConfig
import copy
import six
from .upstream.loadwsgi import (
    APP, FILTER, SERVER, PIPELINE, FILTER_APP, FILTER_WITH, ConfigLoader, loadcontext)

SCHEMEMAP = {
    'application': 'application',
    'app': 'application',
    'composite': 'composite',
    'composit': 'composite',
    'server': 'server',
    'filter': 'filter',
    'filter-app': 'filter-app',
    'pipeline': 'pipeline'
}

TYPEMAP = {
    APP: 'app',
    SERVER: 'server',
    FILTER: 'filter',
}


@attributes(['path'], apply_with_init=False, apply_immutable=True)
@implementer(IConfigLoader)
class PasteDeployConfigLoader(object):
    directlyProvides(IConfigLoaderFactory)

    def __init__(self, path):
        self.path = path
        self.contexts = {}

    def config(self):
        # we can't get everything because globals can be different for different
        # contexts, and montague can't handle that.
        raise NotImplementedError('PasteDeploy does not provide the complete config')

    def _preprocess_name(self, name):
        if name is not None and name.startswith('_montague_pastedeploy'):
            return name.split('_')[3]
        return name

    def _walk_contexts(self, name_parts, context, stack):
        node_name = '_'.join(name_parts)
        if not (hasattr(context, 'filter_context') or hasattr(context, 'next_context')):
            # leaf node
            stack.append((node_name, context))
        else:
            if hasattr(context, 'filter_context'):
                self._walk_contexts(name_parts + ('filter',), context.filter_context, stack)
            if hasattr(context, 'next_context'):
                self._walk_contexts(name_parts + ('next',), context.next_context, stack)

    def _populate_contexts(self, orig_name, context):
        name = orig_name if orig_name is not None else 'main'
        new_orig_name_parts = ('_montague_pastedeploy', name)
        stack = []
        self._walk_contexts(new_orig_name_parts, context, stack)
        outer_context_name = None
        for node_name, context in stack:
            self.contexts[node_name] = {
                'filter-with': outer_context_name,
                'context': context
            }
            outer_context_name = node_name
        return node_name

    def _postprocess_context(self, name):
        context_data = self.contexts[name]
        context_data['context'].local_conf.pop('filter-with', None)
        context_data['context'].local_conf.pop('next', None)
        if context_data['filter-with'] is not None:
            context_data['context'].local_conf['filter-with'] = context_data['filter-with']
        context = context_data['context']
        local_conf = {}
        local_conf.update(context.local_conf)
        use_conf = local_conf.pop('_montague_use')
        local_conf.update(use_conf)
        # if context.object_type not in (FILTER_WITH, FILTER_APP):
        #    raise ValueError('pdb here')

        assert context.object_type in TYPEMAP
        # Sending a loadable_type of 'app' for composite
        # will break when it's a 'call' use type instead of 'egg'
        # TODO: write a test to catch this
        config = LoadableConfig(
            name=name, config=local_conf,
            global_config=context.global_conf,
            loadable_type=TYPEMAP[context.object_type],
            entry_point_groups=[context.protocol])
        return config

    def app_config(self, name):
        uri = 'config:{}'.format(self.path)
        if name is DEFAULT:
            name = None
        orig_name = self._preprocess_name(name)
        if orig_name == name:
            context = loadcontext(APP, uri, name)
            name = self._populate_contexts(orig_name, context)
        return self._postprocess_context(name)

    def server_config(self, name):
        uri = 'config:{}'.format(self.path)
        if name is DEFAULT:
            name = None
        orig_name = self._preprocess_name(name)
        if orig_name == name:
            context = loadcontext(SERVER, uri, name)
            name = self._populate_contexts(orig_name, context)
        return self._postprocess_context(name)

    def filter_config(self, name):
        uri = 'config:{}'.format(self.path)
        if name is DEFAULT:
            name = None
        orig_name = self._preprocess_name(name)
        if orig_name == name:
            context = loadcontext(FILTER, uri, name)
            name = self._populate_contexts(orig_name, context)
        return self._postprocess_context(name)



import os
from montague_pastedeploy.upstream import loadapp

here = os.path.dirname(__file__)


def test_main(fakeapp):
    app = loadapp('config:sample_configs/basic_app.ini',
                  relative_to=here)
    assert app is fakeapp.apps.basic_app
    app = loadapp('config:sample_configs/basic_app.ini#main',
                  relative_to=here)
    assert app is fakeapp.apps.basic_app
    app = loadapp('config:sample_configs/basic_app.ini',
                  relative_to=here, name='main')
    assert app is fakeapp.apps.basic_app
    app = loadapp('config:sample_configs/basic_app.ini#ignored',
                  relative_to=here, name='main')
    assert app is fakeapp.apps.basic_app


def test_other(fakeapp):
    app = loadapp('config:sample_configs/basic_app.ini#other',
                  relative_to=here)
    assert app is fakeapp.apps.basic_app2


def test_composit(fakeapp):
    app = loadapp('config:sample_configs/basic_app.ini#remote_addr',
                  relative_to=here)
    assert isinstance(app, fakeapp.apps.RemoteAddrDispatch)
    assert app.map['127.0.0.1'] is fakeapp.apps.basic_app
    assert app.map['0.0.0.0'] is fakeapp.apps.basic_app2



import os
from montague_pastedeploy.upstream import loadapp


here = os.path.dirname(__file__)


def test_filter_app(fakeapp):
    app = loadapp('config:sample_configs/test_filter.ini#filt',
                  relative_to=here)
    assert isinstance(app, fakeapp.apps.CapFilter)
    assert app.app is fakeapp.apps.basic_app
    assert app.method_to_call == 'lower'


def test_pipeline(fakeapp):
    app = loadapp('config:sample_configs/test_filter.ini#piped',
                  relative_to=here)
    assert isinstance(app, fakeapp.apps.CapFilter)
    assert app.app is fakeapp.apps.basic_app
    assert app.method_to_call == 'upper'


def test_filter_app2(fakeapp):
    app = loadapp('config:sample_configs/test_filter.ini#filt2',
                  relative_to=here)
    assert isinstance(app, fakeapp.apps.CapFilter)
    assert app.app is fakeapp.apps.basic_app
    assert app.method_to_call == 'lower'


def test_pipeline2(fakeapp):
    app = loadapp('config:sample_configs/test_filter.ini#piped2',
                  relative_to=here)
    assert isinstance(app, fakeapp.apps.CapFilter)
    assert app.app is fakeapp.apps.basic_app
    assert app.method_to_call == 'upper'


def test_filter_app_inverted(fakeapp):
    app = loadapp('config:sample_configs/test_filter.ini#inv',
                  relative_to=here)
    assert isinstance(app, fakeapp.apps.CapFilter)
    assert app.app is fakeapp.apps.basic_app


def test_filter_with_filter_with(fakeapp):
    app = loadapp('config:sample_configs/test_filter_with.ini',
                  relative_to=here)
    assert isinstance(app, fakeapp.apps.CapFilter)
    assert isinstance(app.app, fakeapp.apps.CapFilter)
    assert app.app.app is fakeapp.apps.basic_app



# Servers
import webtest


class TestServer(object):
    def __init__(self, global_conf, local_conf):
        self.global_conf = global_conf
        self.local_conf = local_conf

    def __call__(self, wsgi_app):
        test_app = webtest.TestApp(wsgi_app)
        test_app.montague_conf = dict(global_conf=self.global_conf,
                                      local_conf=self.local_conf)
        return test_app


def make_server_factory(global_conf, **local_conf):
    return TestServer(global_conf, local_conf)


def make_server_runner(wsgi_app, global_conf, **local_conf):
    return make_server_factory(global_conf, **local_conf)(wsgi_app)



# (c) 2005 Ian Bicking and contributors; written for Paste (http://pythonpaste.org)
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php
"""Paste Configuration Middleware and Objects"""
import re

__all__ = ['ConfigMiddleware', 'PrefixMiddleware']


class ConfigMiddleware(object):

    """
    A WSGI middleware that adds a ``paste.config`` key to the request
    environment.
    """

    def __init__(self, application, config):
        """
        This delegates all requests to `application`, adding a *copy*
        of the configuration `config`.
        """
        self.application = application
        self.config = config

    def __call__(self, environ, start_response):
        popped_config = None
        if 'paste.config' in environ:
            popped_config = environ['paste.config']
        environ['paste.config'] = self.config.copy()
        app_iter = None
        try:
            app_iter = self.application(environ, start_response)
        finally:
            if app_iter is None:
                # An error occurred...
                if popped_config is not None:
                    environ['paste.config'] = popped_config
        if type(app_iter) in (list, tuple):
            # Because it is a concrete iterator (not a generator) we
            # know the configuration for this request is no longer
            # needed:
            if popped_config is not None:
                environ['paste.config'] = popped_config
        return app_iter


def make_config_filter(app, global_conf, **local_conf):
    conf = global_conf.copy()
    conf.update(local_conf)
    return ConfigMiddleware(app, conf)

make_config_middleware = ConfigMiddleware.__doc__


class PrefixMiddleware(object):
    """Translate a given prefix into a SCRIPT_NAME for the filtered
    application.

    PrefixMiddleware provides a way to manually override the root prefix
    (SCRIPT_NAME) of your application for certain, rare situations.

    When running an application under a prefix (such as '/james') in
    FastCGI/apache, the SCRIPT_NAME environment variable is automatically
    set to to the appropriate value: '/james'. Pylons' URL generating
    functions, such as url_for, always take the SCRIPT_NAME value into account.

    One situation where PrefixMiddleware is required is when an application
    is accessed via a reverse proxy with a prefix. The application is accessed
    through the reverse proxy via the the URL prefix '/james', whereas the
    reverse proxy forwards those requests to the application at the prefix '/'.

    The reverse proxy, being an entirely separate web server, has no way of
    specifying the SCRIPT_NAME variable; it must be manually set by a
    PrefixMiddleware instance. Without setting SCRIPT_NAME, url_for will
    generate URLs such as: '/purchase_orders/1', when it should be
    generating: '/james/purchase_orders/1'.

    To filter your application through a PrefixMiddleware instance, add the
    following to the '[app:main]' section of your .ini file:

    .. code-block:: ini

        filter-with = proxy-prefix

        [filter:proxy-prefix]
        use = egg:PasteDeploy#prefix
        prefix = /james

    The name ``proxy-prefix`` simply acts as an identifier of the filter
    section; feel free to rename it.

    Also, unless disabled, the ``X-Forwarded-Server`` header will be
    translated to the ``Host`` header, for cases when that header is
    lost in the proxying.  Also ``X-Forwarded-Host``,
    ``X-Forwarded-Scheme``, and ``X-Forwarded-Proto`` are translated.

    If ``force_port`` is set, SERVER_PORT and HTTP_HOST will be
    rewritten with the given port.  You can use a number, string (like
    '80') or the empty string (whatever is the default port for the
    scheme).  This is useful in situations where there is port
    forwarding going on, and the server believes itself to be on a
    different port than what the outside world sees.

    You can also use ``scheme`` to explicitly set the scheme (like
    ``scheme = https``).
    """
    def __init__(self, app, global_conf=None, prefix='/',
                 translate_forwarded_server=True,
                 force_port=None, scheme=None):
        self.app = app
        self.prefix = prefix.rstrip('/')
        self.translate_forwarded_server = translate_forwarded_server
        self.regprefix = re.compile("^%s(.*)$" % self.prefix)
        self.force_port = force_port
        self.scheme = scheme

    def __call__(self, environ, start_response):
        url = environ['PATH_INFO']
        url = re.sub(self.regprefix, r'\1', url)
        if not url:
            url = '/'
        environ['PATH_INFO'] = url
        environ['SCRIPT_NAME'] = self.prefix
        if self.translate_forwarded_server:
            if 'HTTP_X_FORWARDED_SERVER' in environ:
                environ['SERVER_NAME'] = environ['HTTP_HOST'] = \
                    environ.pop('HTTP_X_FORWARDED_SERVER').split(',')[0]
            if 'HTTP_X_FORWARDED_HOST' in environ:
                environ['HTTP_HOST'] = environ.pop('HTTP_X_FORWARDED_HOST').split(',')[0]
            if 'HTTP_X_FORWARDED_FOR' in environ:
                environ['REMOTE_ADDR'] = \
                    environ.pop('HTTP_X_FORWARDED_FOR').split(',')[0]
            if 'HTTP_X_FORWARDED_SCHEME' in environ:
                environ['wsgi.url_scheme'] = environ.pop('HTTP_X_FORWARDED_SCHEME')
            elif 'HTTP_X_FORWARDED_PROTO' in environ:
                environ['wsgi.url_scheme'] = environ.pop('HTTP_X_FORWARDED_PROTO')
        if self.force_port is not None:
            host = environ.get('HTTP_HOST', '').split(':', 1)[0]
            if self.force_port:
                host = '%s:%s' % (host, self.force_port)
                environ['SERVER_PORT'] = str(self.force_port)
            else:
                if environ['wsgi.url_scheme'] == 'http':
                    port = '80'
                else:
                    port = '443'
                environ['SERVER_PORT'] = port
            environ['HTTP_HOST'] = host
        if self.scheme is not None:
            environ['wsgi.url_scheme'] = self.scheme
        return self.app(environ, start_response)


def make_prefix_middleware(
        app, global_conf, prefix='/',
        translate_forwarded_server=True,
        force_port=None, scheme=None):
    from .converters import asbool
    translate_forwarded_server = asbool(translate_forwarded_server)
    return PrefixMiddleware(
        app, prefix=prefix,
        translate_forwarded_server=translate_forwarded_server,
        force_port=force_port, scheme=scheme)

make_prefix_middleware.__doc__ = PrefixMiddleware.__doc__



import os
from montague import load_app, load_server, load_filter
from montague.loadwsgi import Loader
from montague.structs import ComposedFilter

here = os.path.dirname(__file__)


def test_load_app(fakeapp):
    config_path = os.path.join(here, 'config_files/simple_config.ini')
    app = load_app(config_path)
    assert app is fakeapp.apps.basic_app


def test_load_server(fakeapp):
    config_path = os.path.join(here, 'config_files/simple_config.ini')
    server = load_server(config_path, name='server_factory')
    actual = server(fakeapp.apps.basic_app)
    assert actual.montague_conf['local_conf']['port'] == '42'
    resp = actual.get('/')
    assert b'This is basic app' == resp.body
    server = load_server(config_path, name='server_runner')
    actual = server(fakeapp.apps.basic_app2)
    assert actual.montague_conf['local_conf']['host'] == '127.0.0.1'
    resp = actual.get('/')
    assert b'This is basic app2' == resp.body


def test_load_filter(fakeapp):
    config_path = os.path.join(here, 'config_files/simple_config.ini')
    filter = load_filter(config_path, name='filter')
    app = filter(None)
    assert isinstance(app, fakeapp.apps.CapFilter)


def test_load_filtered_app(fakeapp):
    config_path = os.path.join(here, 'config_files/simple_config.ini')
    app = load_app(config_path, name='filtered-app')
    assert isinstance(app, fakeapp.apps.CapFilter)
    assert app.app is fakeapp.apps.basic_app
    assert app.method_to_call == 'lower'


def test_load_layered_filter(fakeapp):
    config_path = os.path.join(here, 'config_files/simple_config.ini')
    filter = load_filter(config_path, name='filter1')
    assert isinstance(filter, ComposedFilter)
    app = filter(fakeapp.apps.basic_app)
    assert app.app.app is fakeapp.apps.basic_app
    assert isinstance(app, fakeapp.apps.CapFilter)
    assert isinstance(app.app, fakeapp.apps.CapFilter)


def test_filter_app(fakeapp):
    # Specifically the 'filter-app' config type
    config_path = os.path.join(here, 'config_files/filter_app.ini')
    app = load_app(config_path)
    assert app.method_to_call == 'lower'
    assert app.app.method_to_call == 'upper'
    assert app.app.app is fakeapp.apps.basic_app


def test_nested(fakeapp):
    config_path = os.path.join(here, 'config_files/even_more_nested.ini')
    loader = Loader(config_path)
    loadable = loader._load_app('filt').normalize()
    chain = ('lower', 'swapcase', 'title', 'swapcase', 'reverse', 'upper')
    for item in chain:
        assert loadable.local_conf['method_to_call'] == item
        loadable = loadable.inner
    assert loadable.is_app



from montague_pastedeploy.upstream.config import ConfigMiddleware
import pytest
from webtest import TestApp as WebtestApp


class Bug(Exception):
    pass


def app_with_exception(environ, start_response):
    def cont():
        yield b"something"
        raise Bug
    start_response(b'200 OK', [(b'Content-type', b'text/html')])
    return cont()


def test_error():
    wrapped = ConfigMiddleware(app_with_exception, {'test': 1})
    test_app = WebtestApp(wrapped)
    with pytest.raises(Bug):
        test_app.get('/')



from montague import load_app as montague_loadapp
from paste.deploy import loadapp as paste_loadapp
import os


here = os.path.dirname(__file__)


def test_main(fakeapp):
    app = paste_loadapp('config:sample_configs/basic_app.ini',
                        relative_to=here)
    assert app is fakeapp.apps.basic_app
    app = montague_loadapp(os.path.join(here, 'sample_configs/basic_app.ini'))
    assert app is fakeapp.apps.basic_app

    app = paste_loadapp('config:sample_configs/basic_app.ini',
                        relative_to=here, name='main')
    assert app is fakeapp.apps.basic_app
    app = montague_loadapp(os.path.join(here, 'sample_configs/basic_app.ini'),
                           name='main')
    assert app is fakeapp.apps.basic_app


def test_other(fakeapp):
    app = paste_loadapp('config:sample_configs/basic_app.ini',
                        relative_to=here, name='other')
    assert app is fakeapp.apps.basic_app2
    app = montague_loadapp(os.path.join(here, 'sample_configs/basic_app.ini'),
                           name='other')
    assert app is fakeapp.apps.basic_app2


def test_composit(fakeapp):
    app = paste_loadapp('config:sample_configs/basic_app.ini',
                        relative_to=here, name='remote_addr')
    assert isinstance(app, fakeapp.apps.RemoteAddrDispatch)
    assert app.map['127.0.0.1'] is fakeapp.apps.basic_app
    assert app.map['0.0.0.0'] is fakeapp.apps.basic_app2
    app = montague_loadapp(os.path.join(here, 'sample_configs/basic_app.ini'),
                           name='remote_addr')
    assert isinstance(app, fakeapp.apps.RemoteAddrDispatch)
    assert app.map['127.0.0.1'] is fakeapp.apps.basic_app
    assert app.map['0.0.0.0'] is fakeapp.apps.basic_app2



from montague_pastedeploy.upstream.converters import asbool

def test_asbool_truthy():
    assert asbool('true')
    assert asbool('yes')
    assert asbool('on')
    assert asbool('y')
    assert asbool('t')
    assert asbool('1')

    
def test_asbool_falsy():
    assert not asbool('false')
    assert not asbool('no')
    assert not asbool('off')
    assert not asbool('n')
    assert not asbool('f')
    assert not asbool('0')



from __future__ import absolute_import
# (c) 2005 Ian Bicking and contributors; written for Paste (http://pythonpaste.org)
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php
from .loadwsgi import loadapp, loadserver, loadfilter, appconfig

__all__ = ['loadapp', 'loadserver', 'loadfilter', 'appconfig']



from montague import load_app as montague_loadapp
from paste.deploy import loadapp as paste_loadapp
from paste.deploy import appconfig as paste_appconfig
import os
import pytest

ini_file = 'config:sample_configs/test_config.ini'
here = os.path.dirname(__file__)
config_path = os.path.join(here, 'sample_configs')
config_filename = os.path.join(config_path, 'test_config.ini')


def test_config1(fakeapp):
    paste_app = paste_loadapp(ini_file, relative_to=here, name='test1')
    montague_app = montague_loadapp(config_filename, name='test1')
    for app in (paste_app, montague_app):
        assert app.local_conf == {
            'setting1': 'foo',
            'setting2': 'bar',
            'apppath': os.path.join(config_path, 'app')}
        assert app.global_conf == {
            'def1': 'a',
            'def2': 'b',
            'basepath': config_path,
            'here': config_path,
            '__file__': config_filename}


def test_config2(fakeapp):
    paste_app = paste_loadapp(ini_file, relative_to=here, name='test2')
    montague_app = montague_loadapp(config_filename, name='test2')
    for app in (paste_app, montague_app):
        assert app.local_conf == {
            'local conf': 'something'}
        assert app.global_conf == {
            'def1': 'test2',
            'def2': 'b',
            'basepath': config_path,
            'another': 'TEST',
            'here': config_path,
            '__file__': config_filename}
    # Run this to make sure the global-conf-modified test2
    # didn't mess up the general global conf
    test_config1(fakeapp)


def test_config3(fakeapp):
    paste_app = paste_loadapp(ini_file, relative_to=here, name='test3')
    montague_app = montague_loadapp(config_filename, name='test3')
    for app in (paste_app, montague_app):
        assert isinstance(app, fakeapp.configapps.SimpleApp)
        assert app.local_conf == {
            'local conf': 'something',
            'another': 'something more\nacross several\nlines'}
        assert app.global_conf == {
            'def1': 'test3',
            'def2': 'b',
            'basepath': config_path,
            'another': 'TEST',
            'here': config_path,
            '__file__': config_filename}
    test_config2(fakeapp)


def test_main(fakeapp):
    app = paste_loadapp('config:test_func.ini',
                        relative_to=config_path)
    assert app is fakeapp.apps.basic_app
    app = paste_loadapp('config:test_func.ini#main',
                        relative_to=config_path)
    assert app is fakeapp.apps.basic_app
    app = paste_loadapp('config:test_func.ini',
                        relative_to=config_path, name='main')
    assert app is fakeapp.apps.basic_app
    app = paste_loadapp('config:test_func.ini#ignored',
                        relative_to=config_path, name='main')
    assert app is fakeapp.apps.basic_app


def test_other(fakeapp):
    app = paste_loadapp('config:test_func.ini#other', relative_to=config_path)
    assert app is fakeapp.apps.basic_app2


def test_composit(fakeapp):
    app = paste_loadapp('config:test_func.ini#remote_addr', relative_to=config_path)
    assert isinstance(app, fakeapp.apps.RemoteAddrDispatch)
    assert app.map['127.0.0.1'] is fakeapp.apps.basic_app
    assert app.map['0.0.0.0'] is fakeapp.apps.basic_app2


def test_foreign_config(fakeapp):
    app = paste_loadapp(ini_file, relative_to=here, name='test_foreign_config')
    assert isinstance(app, fakeapp.configapps.SimpleApp)
    assert app.local_conf == {
        'another': 'FOO',
        'bob': 'your uncle'}
    assert app.global_conf == {
        'def1': 'a',
        'def2': 'from include',
        'def3': 'c',
        'basepath': config_path,
        'glob': 'override',
        'here': config_path,
        '__file__': os.path.join(config_path, 'test_config.ini')}


def test_config_get(fakeapp):
    app = paste_loadapp(ini_file, relative_to=here, name='test_get')
    assert isinstance(app, fakeapp.configapps.SimpleApp)
    assert app.local_conf == {
        'def1': 'a',
        'foo': 'TEST'}
    assert app.global_conf == {
        'def1': 'a',
        'def2': 'TEST',
        'basepath': os.path.join(here, 'sample_configs'),
        'here': config_path,
        '__file__': config_filename}


def test_appconfig(fakeapp):
    conf = paste_appconfig(ini_file, relative_to=here, name='test_get')
    assert conf == {
        'def1': 'a',
        'def2': 'TEST',
        'basepath': os.path.join(here, 'sample_configs'),
        'here': config_path,
        '__file__': config_filename,
        'foo': 'TEST'}
    assert conf.local_conf == {
        'def1': 'a',
        'foo': 'TEST'}
    assert conf.global_conf == {
        'def1': 'a',
        'def2': 'TEST',
        'basepath': os.path.join(here, 'sample_configs'),
        'here': config_path,
        '__file__': config_filename}


def test_appconfig_filter_with(fakeapp):
    conf = paste_appconfig('config:test_filter_with.ini', relative_to=config_path)
    assert conf['example'] == 'test'


def test_global_conf(fakeapp):
    conf = paste_appconfig(ini_file, relative_to=here, name='test_global_conf',
                           global_conf={'def2': 'TEST DEF 2', 'inherit': 'bazbar'})
    assert conf == {
        'def1': 'a',
        # Note that this gets overwritten:
        'def2': 'b',
        'basepath': os.path.join(here, 'sample_configs'),
        'here': config_path,
        'inherit': 'bazbar',
        '__file__': config_filename,
        'test_interp': 'this:bazbar',
        }
    assert conf.local_conf == {
        'test_interp': 'this:bazbar'}


def test_interpolate_exception():
    with pytest.raises(Exception) as excinfo:
        paste_appconfig('config:test_error.ini', relative_to=config_path)
    expected = "Error in file %s" % os.path.join(config_path, 'test_error.ini')
    # _loadconfig 'de-windowsifies' the paths, so we must do the same.
    expected = expected.replace('\\', '/')
    assert expected in excinfo.value.message



############################################################
## Apps
############################################################

def simple_app(response, environ, start_response):
    start_response('200 OK', [('Content-type', 'text/html')])
    return [b'This is ', response]

def basic_app(environ, start_response):
    return simple_app(b'basic app', environ, start_response)

def make_basic_app(global_conf, **conf):
    return basic_app

def basic_app2(environ, start_response):
    return simple_app(b'basic app2', environ, start_response)
    
def make_basic_app2(global_conf, **conf):
    return basic_app2

############################################################
## Composits
############################################################

def make_remote_addr(loader, global_conf, **conf):
    apps = {}
    addrs = {}
    for name, value in conf.items():
        if name.startswith('app.'):
            apps[name[4:]] = loader.get_app(value, global_conf)
        elif name.startswith('addr.'):
            addrs[name[5:]] = value
    dispatcher = RemoteAddrDispatch()
    for name in apps:
        dispatcher.map[addrs[name]] = apps[name]
    return dispatcher

class RemoteAddrDispatch(object):
    def __init__(self, map=None):
        self.map = map or {}

    def __call__(self, environ, start_response):
        addr = environ['REMOTE_ADDR']
        app = self.map.get(addr) or self.map['0.0.0.0']
        return app(environ, start_response)

############################################################
## Filters
############################################################

def make_cap_filter(global_conf, method_to_call='upper'):
    def cap_filter(app):
        return CapFilter(app, global_conf, method_to_call)
    return cap_filter

class CapFilter(object):

    def __init__(self, app, global_conf, method_to_call='upper'):
        self.app = app
        self.method_to_call = method_to_call
        self.global_conf = global_conf

    def __call__(self, environ, start_response):
        app_iter = self.app(environ, start_response)
        for item in app_iter:
            yield getattr(item, self.method_to_call)()
        if hasattr(app_iter, 'close'):
            app_iter.close()

